// tictiactoe.cpp
// mmki08-rb
// Loesung von Sebastian Henke und Yves Radunz

#include <iostream>
#include <string>
#include <vector>
#include <time.h>

//global char constants
const char X = 'X';
const char O = 'O';
const char EMPTY = ' ';
const char TIE = 'T';
const char NO_ONE = 'N';

// function prototypes
int askNumber(std::string question, int high, int low = 0);
inline char opponent(char piece);
inline bool isLegal(const std::vector<char>& board, int move);
void announceWinner(char winner, char computer, char human);
void displayBoard(const std::vector<char>& board);
void instructions();

char winner(const std::vector<char>& board);
int humanMove(const std::vector<char>& board, char human);
int computerMove(std::vector<char> board, char computer);


//main function
int main()
{
  srand ( time(NULL) );
  int move;
  std::vector<char> board(9, EMPTY);

  instructions();

  // Who should start (X)?
  char human = ( (rand()%2 == 0) ? X : O );
  char computer = opponent(human);
  char turn = X;
  displayBoard(board);

  while (winner(board) == NO_ONE)
  {
    if (turn == human) {
      move = humanMove(board, human);
      board[move] = human;
    }
    else {
      move = computerMove(board, computer);
      board[move] = computer;
    }
    displayBoard(board);
    turn = opponent(turn);
  }
  announceWinner(winner(board), computer, human);
  return 0;
}


void instructions()
{
    std::cout << "Make your move by entering a number, 0 - 8. The number\n";
    std::cout << "corresponds to the desired board position, as illustrated:\n\n";
    std::cout << " 1 | 2 | 3\n";
    std::cout << " ---------\n";
    std::cout << " 4 | 5 | 6\n";
    std::cout << " ---------\n";
    std::cout << " 7 | 8 | 9\n\n";
}


int askNumber(std::string question, int high, int low)
{
  int number;
  do {
    std::cout << question << " (" << low << " - " << high << "): ";
    std::cin >> number;
  }
  while (number > high || number < low);
  return number;
}


inline char opponent(char piece)
{
  return( (piece == X) ? O : X );
}


void displayBoard(const std::vector<char>& board)
{
  std::cout << "\n\t" << board[0] << " | " << board[1] << " | " << board[2] << "\n\t" << "---------";
  std::cout << "\n\t" << board[3] << " | " << board[4] << " | " << board[5] << "\n\t" << "---------";
  std::cout << "\n\t" << board[6] << " | " << board[7] << " | " << board[8] << "\n\n";
}


char winner(const std::vector<char>& board)
{
  //all possible winning rows
  const int WINNING_ROWS[8][3] = { {0,1,2},{3,4,5},{6,7,8},{0,3,6},{1,4,7},{2,5,8},{0,4,8},{2,4,6}};
  //if any winning row has three value that are the same (and not EMPTY), we have a winner
  for (int row = 0; row < 8; ++row) {
    if ( (board[WINNING_ROWS[row][0]] != EMPTY) &&
         (board[WINNING_ROWS[row][0]] == board[WINNING_ROWS[row][1]]) &&
         (board[WINNING_ROWS[row][1]] == board[WINNING_ROWS[row][2]]) ) {
      return board[WINNING_ROWS[row][0]];
    }
  }
  //check for a tie (no empty squares left)
  if (count(board.begin(), board.end(), EMPTY) == 0) {
    return TIE;
  }
  // since nobody has won and it isn't a tie, the game ain't over
  return NO_ONE;
}


inline bool isLegal(int move, const std::vector<char>& board)
{
  return (board[move] == EMPTY);
}


int humanMove(const std::vector<char>& board, char human)
{
  int move = askNumber("Where will you move?", board.size(), 1) - 1;
  while ( isLegal(move, board) == false ) {
    std::cout << "\nThat square is already occupied. \n";
    move = askNumber("Where will you move?", board.size(), 1) - 1;
  }
  return move;
}


bool zweiSteine(std::vector<char> board, int feld1, int feld2, int feld3, char spieler) {
       if (board[feld1] == spieler && board[feld2] == spieler && board[feld3] == EMPTY) {
             return true;
       } else {
           return false;
       }
}

/*
  gibt es 2 Steine, welche zu einer Reihe, Spalte oder Diagonale gebildet werden knnen
  Falls ja, dann gib eine Mglichkeit zurck, sonst -1
*/
int anwenden(std::vector<char> board, char spieler) {
       int move = -1;
       
       // Horizontal - erste Reihe
       if (zweiSteine(board,0,1,2,spieler)) {move = 2;}
       if (zweiSteine(board,0,2,1,spieler)) {move = 1;}
       if (zweiSteine(board,1,2,0,spieler)) {move = 0;}

       // Horizontal - zweite Reihe
       if (zweiSteine(board,3,4,5,spieler)) {move = 5;}
       if (zweiSteine(board,3,5,4,spieler)) {move = 4;}
       if (zweiSteine(board,4,5,3,spieler)) {move = 3;}

       // Horizontal - dritte Reihe
       if (zweiSteine(board,6,7,8,spieler)) {move = 8;}
       if (zweiSteine(board,6,8,7,spieler)) {move = 7;}
       if (zweiSteine(board,7,8,6,spieler)) {move = 6;}

       // Vertikal - erste Spalte
       if (zweiSteine(board,0,3,6,spieler)) {move = 6;}
       if (zweiSteine(board,0,6,3,spieler)) {move = 3;}
       if (zweiSteine(board,3,6,0,spieler)) {move = 0;}

       // Vertikal - zweite Spalte
       if (zweiSteine(board,1,4,7,spieler)) {move = 7;}
       if (zweiSteine(board,1,7,4,spieler)) {move = 4;}
       if (zweiSteine(board,4,7,1,spieler)) {move = 1;}

       // Vertikal - erste Spalte
       if (zweiSteine(board,2,5,8,spieler)) {move = 8;}
       if (zweiSteine(board,2,8,5,spieler)) {move = 5;}
       if (zweiSteine(board,5,8,2,spieler)) {move = 2;}

       // die Diagonalen
       if (zweiSteine(board,0,4,8,spieler)) {move = 8;}
       if (zweiSteine(board,0,8,4,spieler)) {move = 4;}
       if (zweiSteine(board,4,8,0,spieler)) {move = 0;}

       if (zweiSteine(board,2,4,6,spieler)) {move = 6;}
       if (zweiSteine(board,2,6,4,spieler)) {move = 4;}
       if (zweiSteine(board,4,6,2,spieler)) {move = 2;}    
       
       return move;
}


int computerMove(std::vector<char> board, char computer)
{
  std::cout << "I take square number ";
// random move
  int move;
  char mensch = opponent(computer);

   // whlen zuflliges freies Feld
      move = int(9.0*rand()/(RAND_MAX));
      while ( isLegal(move, board) == false ) {
            move = int(9.0*rand()/(RAND_MAX));
      }


// und jetzt wird gefiltert
       int help = anwenden(board, computer);
       
       // Siegchance durch bereits 2 Steine --> sofort nutzen
       if (help != -1) {
          std::cout << move+1 << std::endl;
          return move = help;         
       }
       
       // Methode Feld 2 und 4 sind vom Gegner und Feld 6 und 8 von uns, dann whle rechte untere Ecke
       if (board[0] == EMPTY && board[1] == mensch && board[3] == mensch && board[2] == EMPTY && board[6] == EMPTY &&
                    board[8] == EMPTY && board[5] == computer && board[7] == computer) {
          std::cout << 9 << std::endl;
          return move = 8;
       }

       // Methode Feld 6 und 8 sind vom Gegner und Feld 2 und 4 von uns, dann whle linke obere Ecke
       if (board[0] == EMPTY && board[1] == computer && board[3] == computer && board[2] == EMPTY && board[6] == EMPTY &&
                    board[8] == EMPTY && board[5] == mensch && board[7] == mensch) {
          std::cout << 1 << std::endl;
          return move = 0;
       }

       // Methode Feld 2 und 6 sind vom Gegner und Feld 4 und 8 von uns, dann whle linke untere Ecke
       if (board[2] == EMPTY && board[1] == mensch && board[5] == mensch && board[0] == EMPTY && board[8] == EMPTY &&
                    board[6] == EMPTY && board[7] == computer && board[3] == computer) {
          std::cout << 7 << std::endl;
          return move = 6;
       }

       // Methode Feld 4 und 8 sind vom Gegner und Feld 2 und 6 von uns, dann whle rechte untere Ecke
       if (board[2] == EMPTY && board[1] == computer && board[5] == computer && board[0] == EMPTY && board[8] == EMPTY &&
                    board[6] == EMPTY && board[7] == mensch && board[3] == mensch) {
          std::cout << 399 << std::endl;
          return move = 2;
       }
       
       // Gefahr durch 2 Steine vom Gegner
       // Siegchance vom Gegner verhindern
       help = anwenden(board, mensch);
       if (help != -1) {
          move = help;         
       }

       // Gegner baut Falle auf (uere Mittelfelder)
       if (board[0] == EMPTY && board[1] == mensch && board[3] == mensch && board[2] == EMPTY && board[6] == EMPTY && help == -1) {
          std::cout << 1 << std::endl;
          return move = 0;
       }
       if (board[2] == EMPTY && board[1] == mensch && board[5] == mensch && board[8] == EMPTY && board[0] == EMPTY && help == -1) {
          std::cout << 3 << std::endl;
          return move = 2;
       }
       if (board[6] == EMPTY && board[3] == mensch && board[7] == mensch && board[0] == EMPTY && board[8] == EMPTY && help == -1) {
          std::cout << 7 << std::endl;
          return move = 6;
       }
       if (board[8] == EMPTY && board[7] == mensch && board[5] == mensch && board[2] == EMPTY && board[6] == EMPTY && help == -1) {
          std::cout << 9 << std::endl;
          return move = 8;
       }


       // Gegner hat seinen ersten Zug in eine Ecke gelegt, dann whle die Mitte
       if (board[0] == mensch && board[1] != mensch && board[2] != mensch && 
                    board[3] != mensch && board[5] != mensch && board[6] != mensch &&
                    board[7] != mensch && board[8] != mensch && board[4] == EMPTY) {
          std::cout << 5 << std::endl;
          return move = 4;
       }   
       if (board[0] != mensch && board[1] != mensch && board[2] == mensch && 
                    board[3] != mensch && board[5] != mensch && board[6] != mensch &&
                    board[7] != mensch && board[8] != mensch && board[4] == EMPTY) {
          std::cout << 5 << std::endl;
          return move = 4;
       }   
       if (board[0] != mensch && board[1] != mensch && board[2] != mensch && 
                    board[3] != mensch && board[5] != mensch && board[6] == mensch &&
                    board[7] != mensch && board[8] != mensch && board[4] == EMPTY) {
          std::cout << 5 << std::endl;
          return move = 4;
       }   
       if (board[0] != mensch && board[1] != mensch && board[2] != mensch && 
                    board[3] != mensch && board[5] != mensch && board[6] != mensch &&
                    board[7] != mensch && board[8] == mensch && board[4] == EMPTY) {
          std::cout << 5 << std::endl;
          return move = 4;
       }   


       // Gegner hat seinen ersten Zug in ein mittleres Auenfeld gelegt, dann whle die Mitte
       if (board[0] != mensch && board[1] == mensch && board[2] != mensch && 
                    board[3] != mensch && board[5] != mensch && board[6] != mensch &&
                    board[7] != mensch && board[8] != mensch && board[4] == EMPTY) {
          std::cout << 5 << std::endl;
          return move = 4;
       }   
       if (board[0] != mensch && board[1] != mensch && board[2] != mensch && 
                    board[3] == mensch && board[5] != mensch && board[6] != mensch &&
                    board[7] != mensch && board[8] != mensch && board[4] == EMPTY) {
          std::cout << 5 << std::endl;
          return move = 4;
       }   
       if (board[0] != mensch && board[1] != mensch && board[2] != mensch && 
                    board[3] != mensch && board[5] == mensch && board[6] != mensch &&
                    board[7] != mensch && board[8] != mensch && board[4] == EMPTY) {
          std::cout << 5 << std::endl;
          return move = 4;
       }   
       if (board[0] != mensch && board[1] != mensch && board[2] != mensch && 
                    board[3] != mensch && board[5] != mensch && board[6] != mensch &&
                    board[7] == mensch && board[8] != mensch && board[4] == EMPTY) {
          std::cout << 5 << std::endl;
          return move = 4;
       }   


       // Gegner auf 2 und 9, Computer in der Mitte
       if (board[0] != mensch && board[1] == mensch && board[2] == EMPTY && 
                    board[3] != mensch && board[5] != mensch && board[6] != mensch &&
                    board[7] != mensch && board[8] == mensch && board[4] == computer) {
          std::cout << 3 << std::endl;
          return move = 2;
       }   
       // Gegner auf 2 und 7, Computer in der Mitte
       if (board[0] == EMPTY && board[1] == mensch && board[2] != mensch && 
                    board[3] != mensch && board[5] != mensch && board[6] == mensch &&
                    board[7] != mensch && board[8] != mensch && board[4] == computer) {
          std::cout << 1 << std::endl;
          return move = 0;
       }   
       // Gegner auf 4 und 3, Computer in der Mitte
       if (board[0] != mensch && board[1] == EMPTY && board[2] == mensch && 
                    board[3] == mensch && board[5] != mensch && board[6] != mensch &&
                    board[7] != mensch && board[8] != mensch && board[4] == computer) {
          std::cout << 2 << std::endl;
          return move = 1;
       }   
       // Gegner auf 3 und 9, Computer in der Mitte
       if (board[0] != mensch && board[1] != mensch && board[2] != mensch && 
                    board[3] == mensch && board[5] != mensch && board[6] == mensch &&
                    board[7] == EMPTY && board[8] == mensch && board[4] == computer) {
          std::cout << 8 << std::endl;
          return move = 7;
       }   
       // Gegner auf 6 und 1, Computer in der Mitte
       if (board[0] == mensch && board[1] == EMPTY && board[2] != mensch && 
                    board[3] != mensch && board[5] == mensch && board[6] != mensch &&
                    board[7] != mensch && board[8] != mensch && board[4] == computer) {
          std::cout << 2 << std::endl;
          return move = 1;
       }   
       // Gegner auf 6 und 7, Computer in der Mitte
       if (board[0] != mensch && board[1] != mensch && board[2] != mensch && 
                    board[3] != mensch && board[5] == mensch && board[6] == mensch &&
                    board[7] == EMPTY && board[8] != mensch && board[4] == computer) {
          std::cout << 8 << std::endl;
          return move = 7;
       }   
       // Gegner auf 8 und 1, Computer in der Mitte
       if (board[0] == mensch && board[1] != mensch && board[2] != mensch && 
                    board[3] == EMPTY && board[5] != mensch && board[6] != mensch &&
                    board[7] == mensch && board[8] != mensch && board[4] == computer) {
          std::cout << 4 << std::endl;
          return move = 3;
       }   
       // Gegner auf 8 und 3, Computer in der Mitte
       if (board[0] != mensch && board[1] != mensch && board[2] == mensch && 
                    board[3] != mensch && board[5] == EMPTY && board[6] != mensch &&
                    board[7] == mensch && board[8] != mensch && board[4] == computer) {
          std::cout << 6 << std::endl;
          return move = 5;
       }   

       // Gegner in der Mitte, dann gehe in die obere linke Ecke
       if (board[0] == EMPTY && board[1] == EMPTY && board[2] == EMPTY && 
                    board[3] == EMPTY && board[5] == EMPTY && board[6] == EMPTY &&
                    board[7] == EMPTY && board[8] == EMPTY && board[4] == mensch) {
          std::cout << 1 << std::endl;
          return move = 0;
       }   


       // Gegner in Ecke und gegenberliegender Ecke und Computer inner Mitte, dann gehe nicht in Ecke
       if (board[0] == mensch && board[1] != mensch && board[2] != mensch && 
                    board[3] != mensch && board[5] != mensch && board[6] != mensch &&
                    board[7] != mensch && board[8] == mensch && board[4] != mensch) {
          std::cout << 2 << std::endl;
          return move = 1;
       }   
       if (board[0] != mensch && board[1] != mensch && board[2] == mensch && 
                    board[3] != mensch && board[5] != mensch && board[6] == mensch &&
                    board[7] != mensch && board[8] != mensch && board[4] != mensch) {
          std::cout << 2 << std::endl;
          return move = 1;
       }   

  std::cout << move+1 << std::endl;
  return (move);
}


void announceWinner(char winner, char computer, char human)
{
  if (winner == computer) {
    std::cout << "Computer has won!\n";
  }
  else if (winner == human) {
    std::cout << "Human has won!\n";
  }
  else {
    std::cout << "It's a tie.\n";
  }
}
