/*
 * Serie4.cpp
 * mmki08-rb
 */



#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <cmath>
#include <time.h>  // fuer srand(time(NULL))
#include <cstdlib> // fuer MAX_RAND



using namespace std;



//global char constants
const char X      = 'X';
const char O      = 'O';
const char EMPTY  = ' ';
const char TIE    = 'T';
const char NO_ONE = 'N';



int MOVES[9] = {0, 1, 2, 3, 4, 5, 6, 7, 8};



typedef map<pair<string, int>, double> actionValueFunction;



// function prototypes
void   displayBoard       (const string&         board                              );
char   winner             (const string&         board                              );
bool   illegalWin         (const string&         board                              );
bool   isValid            (const string&         board                              );
int    win                (const string&         board,    char player              );
void   announceWinner     (      char            winner,   char computer, char human);
string intToBoard         (      int             num                                );
int    computerMoveOptimal(const string&         board,    char computer            );
int    computerMoveRnd    (const string&         board,    char player              );
int    generateStateList  (      vector<string>& stateList                          );
double reward             (const string&         board,    char player              );



// inlines
inline char opponent(char piece) { return ((piece == X) ? O : X); }
inline bool isLegal(int move, const string& board) { return ((move >= 0) && (move <= 8) && (board[move] == EMPTY)); }



// "neue" Funktionen - fuer Aufgabe 4 ggf zu implementieren!
void initQFunction(const vector<string> stateList, actionValueFunction& qValue, double min, double max);
void q_update(actionValueFunction& qValue, const string& oldState, const int action,
              const double reward, const string& newState, const int newAction, double alpha, double gamma);
int    greedy(const string& board, char player, actionValueFunction& qValue, double epsilon);
int  e_greedy(const string& board, char player, actionValueFunction& qValue, double epsilon);

void experiment(const vector<string> stateList, actionValueFunction& qValue, int l, int t, int z, double a, double e, double g, double minV, double maxV,
		int (*pi_learn)(const string& board, char player), int (*pi_test)(const string& board, char player),
		int (*ag_learn)(const string& board, char player, actionValueFunction& qValue, double epsilon), 
		double (*rFkt)(const string& board, char player));




//main function
int main() {
  srand(time(NULL));

  vector<string>      stateList;    // Liste mit den gueltigen Zustaenden
  actionValueFunction qFunction;   // Wertfunktion (Zustand,Aktion -> double)
  generateStateList(stateList);   // alle gueltigen Zustaende erzeugen und in stateList eintragen

  // Lern- und Experimentalparameter setzen
  const double ALPHA   =    0.2;
  const double EPSILON =    0.2;
  const double GAMMA   =    1.0;
  const int    L       = 2000; // 2000
  const int    T       =   10; // 1
  const int    Z       =   20; // 80

  int lost = 0;
  int tie  = 0;
  int won  = 0;

  cout << "\n|   --   Agent has won.\n.   --   Tie.\n-   --   Opponent has won.\n";

  cout << "\n\n==== EXPERIMENT 4d (Sarsa) ====\n";
  experiment(stateList, qFunction, L, T, Z, ALPHA, EPSILON, GAMMA, -1.0, 1.0, &computerMoveOptimal, &computerMoveOptimal, &e_greedy, &reward);

  cout << "\n\n=== EXPERIMENT 4d (Watkins) ===\n";
  experiment(stateList, qFunction, L, T, Z, ALPHA, EPSILON, GAMMA, -1.0, 1.0, &computerMoveOptimal, &computerMoveOptimal, &greedy, &reward);

  cout << "\n\n==== EXPERIMENT 4e (Sarsa) ====\n";
  experiment(stateList, qFunction, L, T, Z, ALPHA, EPSILON, GAMMA, -1.0, 1.0, &computerMoveRnd, &computerMoveRnd, &e_greedy, &reward);

  cout << "\n\n=== EXPERIMENT 4e (Watkins) ===\n";
  experiment(stateList, qFunction, L, T, Z, ALPHA, EPSILON, GAMMA, -1.0, 1.0, &computerMoveRnd, &computerMoveRnd, &greedy, &reward);

  cout << "\n";
  return 0;
}



void displayBoard(const string& board) {
  cout << "\n\t" << board[0] << " | " << board[1] << " | " << board[2] << "\n\t" << "---------";
  cout << "\n\t" << board[3] << " | " << board[4] << " | " << board[5] << "\n\t" << "---------";
  cout << "\n\t" << board[6] << " | " << board[7] << " | " << board[8] << "\n\n";
}



void announceWinner(char winner, char agent, char opp) {
  if (winner == agent)
    cout << "Agent has won!\n";
  else if (winner == opp)
    cout << "Opponent has won!\n";
  else
    cout << "It's a tie.\n";
}



char winner(const string& board) {
  //all possible winning rows
  const int WINNING_ROWS[8][3] = {{0, 1, 2}, {3, 4, 5}, {6, 7, 8}, {0, 3, 6},
				  {1, 4, 7}, {2, 5, 8}, {0, 4, 8}, {2, 4, 6}};
  //if any winning row has three value that are the same (and not EMPTY), we have a winner
  for (int row = 0; row < 8; ++row)
    if ((board[WINNING_ROWS[row][0]] != EMPTY)
	&& (board[WINNING_ROWS[row][0]] == board[WINNING_ROWS[row][1]])
	&& (board[WINNING_ROWS[row][1]] == board[WINNING_ROWS[row][2]]))
      return board[WINNING_ROWS[row][0]];
  //check for a tie (no empty squares left)
  if (count(board.begin(), board.end(), EMPTY) == 0)
    return TIE;
  // since nobody has won and it isn't a tie, the game ain't over
  return NO_ONE;
}



int win(const string& board, char player ) {
  string newBoard = board;
  for (unsigned int move = 0; move < newBoard.size(); ++move) {
    if (isLegal(move, newBoard)) {
      newBoard[move] = player;
      if (winner(newBoard) == player) {
        return move;
      }
      newBoard[move] = EMPTY; // undo the move
    }
  }
  return -1;
}



int computerMoveOptimal(const string& board, char computer) {
  char human = opponent(computer);

  // REGEL 1: 3-er machen (wenn mglich)
  int winMove = win(board, computer);
  if (winMove != -1)
    return winMove;

  // REGEL 2: 3-er blocken (wenn mglich)
  int blockMove = win(board, human);
  if (blockMove != -1)
    return blockMove;

  // REGEL 3: (nur ntig, wenn Computer mit O spielt)
  if (computer == O) {
    // 3a: wenn der Gegner (X) schon 2 gegenberliegende Ecken hat, dann anlegen
    if (((board[0] == human) && (board[8] == human))
	||((board[2] == human) && (board[6] == human))) {
      const int ANSWER[] = {1, 3, 5, 7};
      for (unsigned int i = 0; i < 4; ++i) {
        int move = ANSWER[i];
        if (isLegal(move, board))
          return move;
      }
    }
    // 3b wenn der Gegner (X) schon 2 gegenberliegende Kanten hat, dann dazwischenlegen
    if (((board[1] == human) || (board[7] == human))
	&& ((board[3] == human) || (board[5] == human))) {
      int move = (board[1] == human ? 1 : 7) + (board[5] == human ? 1 : -1);
      if (isLegal(move, board))
        return move;
    }
  }

  // REGEL 4: Priorittenliste abarbeiten
  // 4a: Zentrum besetzen
  if (isLegal(4, board))
    return 4;
  // 4b: gegenberliegende Ecke besetzen (vom Gegner)
  const int CORNERS[] = {0, 2, 6, 8};
  for (unsigned int i = 0; i < 4; ++i)
    if ((board[CORNERS[i]] == human) && (board[CORNERS[3 - i]] == EMPTY))
      return CORNERS[3-i];
  // 4c: nchste freie Ecke, sonst nchster freier Rand
  const int BEST_MOVES[] = {0, 2, 6, 8, 1, 3, 5, 7};
  for (unsigned int i = 0; i < 8; ++i) {
    int move = BEST_MOVES[i];
    if (isLegal(move, board))
      return move;
  }
  return 42;    //Unreachable statement? Hmm ....
}



bool illegalWin(const string& board) { // checkt, ob evtl. beide Spieler gleichzeitig eine Gewinnreihe haben
  const int WR[8][3] = {{0, 1, 2}, {3, 4, 5}, {6, 7, 8}, {0, 3, 6},
			{1, 4, 7}, {2, 5, 8}, {0, 4, 8}, {2, 4, 6}};
  int c = 0; // Zaehler fuer inkommensurable Gewinnreihen (darf nicht grer als 1 sein)
  char winner;
  for (int i = 0; i < 8; ++i)
    if ((board[WR[i][0]] != EMPTY)
	&& (board[WR[i][0]] == board[WR[i][1]])
	&& (board[WR[i][1]] == board[WR[i][2]]))
      if ((c == 0) || ((c > 0) && (winner != board[WR[i][0]]))) {
        winner = board[WR[i][0]];
        ++c;
      }
  return (c > 1);
}



bool isValid(const string& board) { // checkt ob ein Zustand gueltig ist
  const int x = count(board.begin(), board.end(), X);       // Anz. der X
  const int o = count(board.begin(), board.end(), O);      // Anz. der O
  const int e = count(board.begin(), board.end(), EMPTY); // Anz. der Leerfelder
  return !(illegalWin(board)
	   || ((winner(board) == O) && (x != o))
	   || ((winner(board) == X) && (x != o + 1))
	   || ((e == 0) && (x != o + 1))
	   || ((x < o) || (x > o + 1)));
}



int generateStateList(vector<string>& stateList) {
  const int MAX = int(pow(3.0, 9)) - 1;         // Anzahl alle moeglichen Belegungen des Feldes
  for (int i = 0; i <= MAX; ++i) {
    string board = intToBoard(i);             // Feld generieren
    if (isValid(board))                      // check ob Feld gueltig ist
      stateList.push_back(board);           // Feld in die Zustandsliste eintragen
  }
  return stateList.size();
}



string intToBoard(int num) {
  string BASE;
  ((BASE += EMPTY) += O) += X;
  int i = 8;
  string board(9, EMPTY);
  do {
    board[i--] = BASE[num % 3];
    num /= 3;
  } while(num);
  return board;
}



// ----------------------------------------------------------------------------
int computerMoveRnd(const string& board, char player) {
  // gibt einen zufaelligen gueltigen Zug fuer den Zustand board zurueck
  int move = rand() % 9;
  while (!isLegal(move, board))
    move = rand() % 9;
  return move;
}



double reward(const string& board, char player) {
  // gibt den reward von TTT fuer den angegebenen Zustand und Spieler zurueck
  if (winner(board) == player)
    return 1.0;
  else if (winner(board) == opponent(player))
    return -1.0;
  else 
    return 0.0;
}



void initQFunction(const vector<string> stateList, actionValueFunction& qValue, double min, double max) {
  // die Musterloesung ttt_3l.cpp war hier sicher eine Hilfe
  for (unsigned int i = 0; i < stateList.size(); ++i)
    for (int j = 0; j < 9; j++)
      qValue[pair<string, int>(stateList[i], j)]
	= ((winner(stateList[i]) == X) || (winner(stateList[i]) == O) || (winner(stateList[i]) == TIE)) ? 0 : (double(rand()) / RAND_MAX * (max - min) + min);
}



int greedy(const string& board, char player, actionValueFunction& qValue, double epsilon) {
  /* greedy(s) = argmax_u Q(s,u)
   * die Musterlsung ttt_3l.cpp war auch hier irgendwie eine Hilfe
   * epsilon ist nur zur Kompatibilitaet mit e_greedy vorhanden => unused parameter ^^
   */
  double maxValue;
  int    bestMove;

  /*
   * 1. Suche nach einem Startwert von maxValue, der ganz sicher kleiner ist, als das Maximum der Werte, die in qValue
   *    fuer den aktuellen Zustand gespeichert sind. Schliesslich ist ja nicht gesagt, dass so etwas wie -1000 nicht
   *    doch noch von qValue durch eine ausreichend boese Initialisierung unterboten werden koennte.
   */
  for (unsigned int move = 0; move < 9; ++move)
    if (isLegal(move, board))
      maxValue = qValue[pair<string, int>(board, move)] - 1;

  /*
   * 2. Suche nach dem Maximum.
   */
  for (unsigned int move = 0; move < 9; ++move)
    if (isLegal(move, board)) {
      const double v = qValue[pair<string, int>(board, move)];
      if (v > maxValue) {
        maxValue = v;
        bestMove = move;
      }
    }
  
  /*
   * 3. Zufaellige Auswahl eines Zuges, welcher den maximalen Wert erreicht.
   */
  while(true)
    if (qValue[pair<string, int>(board, bestMove = computerMoveRnd(board, player))] == maxValue)
      return bestMove;

  return (int)(3.141592653589793238*3.141592653589793238/6);     // Unreachable Statement.
}


int e_greedy(const string& board, char player, actionValueFunction& qValue, double epsilon) {
  // Die Musterlsung ttt_3l.cpp gefiel mir nicht.
  if ((double(rand()) / RAND_MAX) < epsilon)      // gibt in epsilon/1 Faellen 
    return computerMoveRnd(board, player);       // einen zufaelligen gueltigen Zug zurueck
  else                                          // sonst
    return greedy(board, player, qValue, 0);   // greedy.
}



void q_update(actionValueFunction& qValue, const string& oldState, const int action, const double reward,
	      const string& newState, const int newAction, double alpha, double gamma) {
  // fuehre ein update auf der Wertfunktion aus (je nach newAction kann dies z.B. SARSA oder WatkinsQ sein)
  
  const double oldQValue = qValue[pair<string, int>(oldState, action)];
  qValue[pair<string, int>(oldState, action)] = oldQValue + alpha * (reward + gamma * qValue[pair<string, int>(newState, newAction)] - oldQValue);
}



void experiment(const vector<string> stateList, actionValueFunction& qValue, int l, int t, int z, double a,
		double e, double g, double minV, double maxV, int (*pi_learn)(const string& board, char player),
		int (*pi_test)(const string& board, char player),
		int (*ag_learn)(const string& board, char player, actionValueFunction& qValue, double epsilon), 
		double (*rFkt)(const string& board, char player)) {
  int lost = 0;
  int tie  = 0;
  int won  = 0;

  cout << "* Experiment Parameters: L = " << l << "  ;  T = " << t << "  ;  Z = "  << z << "\n";
  cout << "* Learning Parameters  : alpha = " << a << "  ;  epsilon = " << e << "  ;  gamma = "  << g << "\n";

  for (int exp = 1; exp <= z; ++exp) {                  // aeussere Experiment-Schleife
    initQFunction(stateList, qValue, -1.0, 1.0);       // Valuefunktion initialisieren

    char   agent = X;                                // Lernagent spielt mit X
    char   opp   = opponent(agent);
    char   turn  = X;                              // X faengt an
    int    ownMove;
    string board(9, EMPTY);                      // leeres Feld erzeugen

    for (int learn = 1; learn <= l; ++learn) { // Lernschleife
      board    = string(9, EMPTY);
      turn = X;

      while (winner(board) == NO_ONE) {     // ab hier wird gespielt
	string oldBoard = board;
	
	oldBoard = board;
	ownMove = e_greedy(board, agent, qValue, e);          // Zuerst darf sich der Agent am Spiel versuchen...
	board[ownMove] = agent;

	if (winner(board) == NO_ONE) {
	  turn = opponent(turn);
	  
	  board[(*pi_learn)(board, opp)] = opp;
	  turn = opponent(turn);
	}
	const int aNew = (winner(board) != NO_ONE ? 0 : (*ag_learn)(board, agent, qValue, e));     // Agent
	q_update(qValue, oldBoard, ownMove, rFkt(board, agent), board, aNew, a, g);          // qFunction aufdatieren
      }
    }
    
    //    for (int i = 0; i < 9; ++i)
    //printf("Aktion: %f\n", qValue[pair<string, int>("         ", i)]);

    for (int test = 1; test <= t; ++test) { //ab hier wird getestet
      int move;
      turn  = X;
      board = string(9, EMPTY);
      while (winner(board) == NO_ONE) {
	if (turn == opp) {
          move = (*pi_test)(board, opp);
	  board[move] = opp;
	} else {
	  move        = greedy(board, agent, qValue, 0);
	  board[move] = agent;
	}
	turn = opponent(turn);
      }
      //      announceWinner(winner(board), agent, opp);
      if (winner(board) == agent) {
	won++;
	cout << "|";
      }
      if (winner(board) == opp) {
	lost++;
	cout << "-";
      }
      if (winner(board) == TIE) {
	tie++;
	cout << ".";
      }
      cout.flush();
    }
  }
  cout << "\n>>> Test Stat : LOST = " << lost << "  |  TIE = " << tie << "  |  WON = "  << won << "\n";  
}
