// ttt_5d.cpp
// mmki08-rb
// Loesung von Sebastian Henke und Yves Radunz

#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <map>
#include <cmath>
#include <time.h>   // fuer srand(time(NULL))
#include <cstdlib> // fuer MAX_RAND
#include <iterator>
#include <fstream>
#include <sstream>

using namespace std;

//global char constants
const char X          = 'X';
const char O          = 'O';
const char EMPTY      = ' ';
const char TIE        = 'T';
const char NO_ONE     = 'N';
const int  BOARD_SIZE = 16;


struct eState{
  char   turn;
  char   winner;
  double value;
};

// function prototypes
int    askNumber            (      string               question,         int                  high,       int                  low                        = 0);
char   askYesNo             (      string               question                                                                                              );
void   instructions         (                                                                                                                                 );
void   displayBoard         (const string&              board                                                                                                 );
char   winner               (const string&              board                                                                                                 );
int    readStatesFromFile   (      vector<string>&      stateList,  const string&              fileName                                                       );
int    readStateInfoFromFile(      map<string, eState>& stateInfo,  const string&              fileName                                                       );
int    greedy               (const string&              board,            char                 player,     map<string, double>& stateValue                    );
int    e_greedy             (const string&              board,            char                 player,     map<string, double>& stateValue, double epsilon    );
int    computerMoveOptimal  (const string&              board,            char                 computer,   map<string, eState>& stateInfo                     );
void   initValueFunction    (const vector<string>       stateList,        map<string, double>& stateValue, double               min,        double max        );
double reward               (const string&              board,            char                 player                                                         );
void   td_update            (      map<string, double>& stateValue, const string&              oldState,
                             const string&              newState,         double               reward,     double               alpha,
                                   double               gamma,            map<string, double>& eTrace,     double               lambda,
                                   vector<string>&      traceList                                                                                             );
void   experiment           (const vector<string>       stateList,        map<string, eState>& stateInfo,  int                  l,          int    t,
                                   int                  z,                double               alpha,      double               eps,        double gamma,
                                   double               lambda                                                                                                );

// inlines
inline char opponent(char piece                    ) { return ((piece == X) ? O : X);                                         }
inline bool isLegal (int  move, const string& board) { return ((move >= 0) && (move < BOARD_SIZE) && (board[move] == EMPTY)); }



//main function
int main() {
  vector<string>      stateList;   // Liste mit den gueltigen Zustaenden
  map<string, eState> stateInfo;  // Liste mit den gueltigen Zustaenden und Bewertungen

  double ALPHA   =  0.1;
  double LAMBDA  =  0.9;
  double EPSILON =  0.05;
  double GAMMA   =  1.0;
  int L          = 50;
  int T          = 10;
  int Z          = 10;

  srand(time(NULL));
  cerr  << "Read " << readStatesFromFile(stateList, "states.txt")       << " board states from \"states.txt\"\n";
  cerr  << "Read " << readStateInfoFromFile(stateInfo, "stateInfo.txt") << " board states from \"stateInfo.txt\"\n";

  experiment(stateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA, LAMBDA);
  experiment(stateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA, 0.0   );

  L = 500;
  experiment(stateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA, LAMBDA);
  experiment(stateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA, 0.0   );

  L = 5000;
  experiment(stateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA, LAMBDA);
  experiment(stateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA, 0.0   );

  L = 50000;
  experiment(stateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA, LAMBDA);
  experiment(stateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA, 0.0   );

  L = 500000;
  experiment(stateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA, LAMBDA);
  experiment(stateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA, 0.0   );

  return 0;
}



void displayBoard(const string& board) {
  cout << "\n\t" << board[ 0] << " | " << board[ 1] << " | " << board[ 2] << " | " << board[ 3] << "\n\t" << "--------------";
  cout << "\n\t" << board[ 4] << " | " << board[ 5] << " | " << board[ 6] << " | " << board[ 7] << "\n\t" << "--------------";
  cout << "\n\t" << board[ 8] << " | " << board[ 9] << " | " << board[10] << " | " << board[11] << "\n\t" << "--------------";
  cout << "\n\t" << board[12] << " | " << board[13] << " | " << board[14] << " | " << board[15] << "\n\n";
}



char winner(const string& board) {
  //all possible winning rows
  const int WINNING_ROWS[10][4] = {{0, 1, 2,  3}, {4, 5,  6,  7}, {8, 9, 10, 11}, {12, 13, 14, 15}, {0, 4, 8, 12},
				   {1, 5, 9, 13}, {2, 6, 10, 14}, {3, 7, 11, 15}, { 0,  5, 10, 15}, {3, 6, 9, 12}};
  for (int row = 0; row < 10; ++row)
    if ((board[WINNING_ROWS[row][0]] != EMPTY)
	&& (board[WINNING_ROWS[row][0]] == board[WINNING_ROWS[row][1]])
	&& (board[WINNING_ROWS[row][0]] == board[WINNING_ROWS[row][2]])
        && (board[WINNING_ROWS[row][0]] == board[WINNING_ROWS[row][3]]))
      return board[WINNING_ROWS[row][0]];

  //check for a tie (no empty squares left)
  if (count(board.begin(), board.end(), EMPTY) == 0)
    return TIE;

  // since nobody has won and it isn't a tie, the game ain't over
  return NO_ONE;
}



int readStateInfoFromFile(map<string, eState>& stateInfo, const string& fileName) {
  fstream stateInfoFile;
  string  line;
  string  board;
  char    turn;
  char    winner;
  double  value;
  stateInfo.clear();
  stateInfoFile.open(fileName.c_str(), ios::in);
  while (!stateInfoFile.eof()) {
    getline(stateInfoFile, line);
    if (line != "") {
      board  = line.substr(0, 16);
      turn   = line[17];
      winner = line[19];
      value  = atof(line.substr(21, line.length() - 21).c_str());
      eState info = {turn, winner, value};
      stateInfo[board] = info;
    }
  }
  stateInfoFile.close();
  return stateInfo.size();
}



int readStatesFromFile(vector<string>& stateList, const string& fileName) {
  fstream statesFile;
  string board;
  stateList.clear();
  statesFile.open(fileName.c_str(), ios::in);
  while (!statesFile.eof())  {
    getline(statesFile, board);
    if (board != "")
      stateList.push_back(board);
  }
  statesFile.close();
  return stateList.size();
}



int greedy(const string& board, char player, map<string, double>& stateValue) {
  string nextBoard         = board;
  double maxValue          = -1000.0;
  int    bestMove          =    -1;
  int    MOVES[BOARD_SIZE] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15};
  random_shuffle(MOVES, MOVES + 9);
  for (int i = 0; i < BOARD_SIZE; ++i) {
    int move = MOVES[i];
    if (isLegal(move, nextBoard)) {
      nextBoard[move] = player;
      const double v = stateValue[nextBoard];
      if (v > maxValue) {
        maxValue = v;
        bestMove = move;
      }
      nextBoard[move] = EMPTY;
    }
  }
  return bestMove;
}



int e_greedy(const string& board, char player, map<string, double>& stateValue, double epsilon) {
  if ((double(rand()) / RAND_MAX) < epsilon) {
    int move = rand() % BOARD_SIZE;
    while(!isLegal(move, board))
      move = rand() % BOARD_SIZE;
    return move;
  }
  return greedy(board, player, stateValue );
}



int computerMoveOptimal(const string& board, char computer, map<string, eState>& stateInfo) {
  double S                 = (computer == X ? 1.0 : -1.0);
  string nextBoard         = board;
  double maxValue          = -1000.0;
  int    bestMove          =    -1;
  int    MOVES[BOARD_SIZE] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15};
  random_shuffle(MOVES, MOVES + BOARD_SIZE);
  for (int i = 0; i < BOARD_SIZE; ++i) {
    int move = MOVES[i];
    if (isLegal(move, nextBoard)) {
      nextBoard[move] = computer;
      const double v = S * stateInfo[nextBoard].value;
      if (v > maxValue) {
        maxValue = v;
        bestMove = move;
      }
      nextBoard[move] = EMPTY;
    }
  }
  return bestMove;
}



void initValueFunction(const vector<string> stateList, map<string, double>& stateValue, double min, double max) {
  for (vector<string>::const_iterator ci = stateList.begin(); ci < stateList.end(); ++ci) {
    const string& board = *ci;
    if (winner(board) != NO_ONE)
      stateValue[board] = 0.0;
    else
      stateValue[board] = double(rand()) / RAND_MAX * (max - min) + min;
  }
}



double reward(const string& board, char player) {
  if (winner(board) == player)
    return 1.0;
  if (winner(board) == opponent(player)) 
    return -1.0;
  return  0.0;
}



void td_update(map<string, double>& stateValue, const string& oldState, const string&         newState, 
	       double               reward,           double  alpha,          double          gamma,
	       map<string, double>& eTrace,           double  lambda,         vector<string>& traceList) {
  // Menge der besuchten Zustaende aufdatieren
  traceList.push_back(oldState);

  // eTrace-Update(1): fuer den gerade besuchten Zustand um 1 vergroessern
  eTrace[oldState] = eTrace[oldState] + 1;

  // stateValue-Update(1): Differenz berechnen
  double d = reward + gamma * stateValue[newState] - stateValue[oldState];

  // eTrace-Update(2):     fuer jeden besuchten Zustand die trace-Variable verkleinern
  // stateValue-Update(2): fuer jeden besuchten Zustand den stateValue-Eintrag aktualisieren
  for (int i = 0; i < traceList.size(); ++i) {
    const string& board = traceList[i];
    eTrace[board]       = eTrace[board] * gamma * lambda;
    stateValue[board]   = stateValue[board] + alpha * d * eTrace[board];
  }
}



void experiment(const vector<string> stateList, map<string, eState>& stateInfo, int    l,
                      int            t,         int                  z,         double alpha,
                      double         eps,       double               gamma,     double lambda) {
  cerr << "\n* Experiment Parameters: L=" << l << "  ;  T=" << t << "  ;  Z="  << z << "\n";
  cerr << "* Learning Parameters  : alpha=" << alpha << "  ;  epsilon=" << eps << "  ;  gamma="  << gamma << "  ;  lambda="  << lambda << "\n";

  int lost = 0;
  int tie  = 0;
  int won  = 0;

  map<string, double> stateValue;   // map mit der Wertfunktion V(s)
  map<string, double> e_trace;     // map mit den trace-Variablen e(s)

  for (int exp = 1; exp <= z; ++exp) { // aeussere Experiment-Schleife
    initValueFunction(stateList, stateValue, 0.0, 0.0);

    char           agent = X;     // Lernagent spielt mit X
    char           opp   = opponent(agent);
    char           turn  = X;   // X faengt an
    string         board(BOARD_SIZE, EMPTY); // leeres Feld erzeugen
    vector<string> traceList;               /* speichert die besuchten Zustaende: Eigentlich eine Menge, aber bei diesem Spiel kann man einen Zustand sowieso
                                             * nie doppelt besuchen. */

    for (int learn = 1; learn <= l; ++learn) {  // Lernschleife
      if (learn % 10000 == 0)
	cerr << ".";
      turn  = X;
      board = string(BOARD_SIZE, EMPTY);
      e_trace.clear();            // eligibility-Traces vor jeder Lernepisode auf 0 setzen!
      traceList.clear();          // analog: die Menge der besuchten Zustaende zuruecksetzen
      while (winner(board) == NO_ONE) { // ab hier wird gespielt
        const string oldBoard = board;
        if (turn == opp) {
          board[computerMoveOptimal(board, opp, stateInfo)] = opp;
          const double r                                    = reward(board, agent);
          td_update(stateValue, oldBoard, board, r, alpha, gamma, e_trace, lambda, traceList);
        } else {
          board[e_greedy(board, agent, stateValue, eps)]    = agent;
          const double r                                    = reward(board, agent);
          td_update(stateValue, oldBoard, board, r, alpha, gamma, e_trace, lambda, traceList);
        }
        turn = opponent(turn);
      }
    }

    for (int test = 1; test <= t; ++test) {  // ab hier wird getestet
      turn  = X;
      board = string(BOARD_SIZE, EMPTY);
      while (winner(board) == NO_ONE) {
        if (turn == opp)
          board[computerMoveOptimal(board, opp, stateInfo)] = opp;
        else
          board[greedy(board, agent, stateValue)] = agent;
        turn = opponent(turn);
      }
      if (winner(board) == agent)
	won++;
      if (winner(board) == opp)
	lost++;
      if (winner(board) == TIE) 
	tie++;
    }
  }
  cerr << "\n>>> Test Stat : LOST = " << lost << "  |  TIE = " << tie << "  |  WON = "  << won << "\n";
}

