  // ttt_5e.cpp
 // mmki08-rb
// Loesung von Sebastian Henke und Yves Radunz



#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <map>
#include <cmath>
#include <time.h>   // fuer srand(time(NULL))
#include <cstdlib> // fuer MAX_RAND
#include <iterator>
#include <fstream>
#include <sstream>



using namespace std;



//global char constants
const char X          = 'X';
const char O          = 'O';
const char EMPTY      = ' ';
const char TIE        = 'T';
const char NO_ONE     = 'N';
const int  BOARD_SIZE = 16;



struct eState{
  char   turn;
  char   winner;
  double value;
};



// function prototypes
char   winner               (const string&              board                                                                                     );
int    readStatesFromFile   (      vector<string>&      stateList,  const string&              fileName                                           );
int    readStateInfoFromFile(      map<string, eState>& stateInfo,  const string&              fileName                                           );
int    greedy               (const string&              board,            char                 player,           map<string, double>& stateValue  );
int    e_greedy             (const string&              board,            char                 player,           map<string, double>& stateValue, 
			           double               epsilon                                                                                   );
int    computerMoveOptimal  (const string&              board,            char                 computer,         map<string, eState>& stateInfo   );
void   initValueFunction    (const vector<string>       stateList,        map<string, double>& stateValue,       double               min, 
			           double               max                                                                                       );
double reward               (const string&              board,            char                 player                                             );
void   td_update            (      map<string, double>& stateValue, const string&              oldState,   const string&              newState,
				   double               reward,           double               alpha,            double               gamma       );
void   experiment           (const vector<string>       stateList,        map<string, eState>& stateInfo,        int                  l,
			           int                  t,                int                  z,                double               alpha,
			           double               eps,              double               gamma                                              );
string altState             (const string&              board                                                                                     );



map<string, string> TRANSFORM;



// inlines
inline char opponent(char piece)                    { return ((piece == X) ? O : X);                               }
inline bool isLegal (int move, const string& board) { return ((move>=0)&&(move<BOARD_SIZE)&&(board[move]==EMPTY)); }



//main function
int main() {
  srand (time(NULL));

  vector<string>      stateList;   // Liste mit den gueltigen Zustaenden
  map<string, eState> stateInfo;  // Liste mit den gueltigen Zustaenden

  int a  = readStatesFromFile( stateList, "states.txt" );
  cerr  << "Read " << a << " board states from \"states.txt\"\n";
  int b  = readStateInfoFromFile( stateInfo, "stateInfo.txt" );
  cerr  << "Read " << b << " board states from \"stateInfo.txt\"\n";


  map<string, double> altStateMap;    // map mit den alternativen Zustaenden (convenience)
  vector<string>      altStateList;  // Liste mit den alternativen Zustaenden

  for (vector<string>::iterator i = stateList.begin(); i != stateList.end(); ++i) {
    altStateMap[altState(*i)] = 0.0;
    TRANSFORM[*i] = altState(*i);
  }

  for (map<string, double>::iterator i = altStateMap.begin(); i != altStateMap.end(); ++i)
    altStateList.push_back( i->first );
  cerr << "Generated " << altStateList.size() << " alternative states.\n";

  double ALPHA   =  0.1;
  double EPSILON =  0.05;
  double GAMMA   =  1.0;
  int    T       = 10;
  int    Z       = 10;

  int L          = 50;
  experiment(altStateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA);

  L = 500;
  experiment(altStateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA);

  L = 5000;
  experiment(altStateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA);

  L = 50000;
  experiment(altStateList, stateInfo, L, T, Z, ALPHA, EPSILON, GAMMA);

  return 0;
}



char winner(const string& board) {
  //all possible winning rows
  const int WINNING_ROWS[10][4] = {{0, 1, 2,  3}, {4, 5,  6,  7}, {8, 9, 10, 11}, {12, 13, 14, 15}, {0, 4, 8, 12},
				   {1, 5, 9, 13}, {2, 6, 10, 14}, {3, 7, 11, 15}, { 0,  5, 10, 15}, {3, 6, 9, 12}};
  for (int row = 0; row < 10; ++row)
    if ((board[WINNING_ROWS[row][0]] != EMPTY)
	&& (board[WINNING_ROWS[row][0]] == board[WINNING_ROWS[row][1]])
	&& (board[WINNING_ROWS[row][0]] == board[WINNING_ROWS[row][2]])
        && (board[WINNING_ROWS[row][0]] == board[WINNING_ROWS[row][3]]))
      return board[WINNING_ROWS[row][0]];

  //check for a tie (no empty squares left)
  if (count(board.begin(), board.end(), EMPTY) == 0)
    return TIE;

  // since nobody has won and it isn't a tie, the game ain't over
  return NO_ONE;
}



int readStateInfoFromFile(map<string, eState>& stateInfo, const string& fileName) {
  fstream stateInfoFile;
  string  line;
  string  board;
  char    turn;
  char    winner;
  double  value;
  stateInfo.clear();
  stateInfoFile.open(fileName.c_str(), ios::in);
  while (!stateInfoFile.eof()) {
    getline( stateInfoFile, line );
    if (line != "") {
      board  = line.substr(0, 16);
      turn   = line[17];
      winner = line[19];
      value  = atof(line.substr(21, line.length()-21).c_str());
      eState info = {turn, winner, value};
      stateInfo[board] = info;
    }
  }
  stateInfoFile.close();
  return stateInfo.size();
}


int readStatesFromFile(vector<string>& stateList, const string& fileName) {
  fstream statesFile;
  string  board;
  stateList.clear();
  statesFile.open(fileName.c_str(), ios::in);
  while (!statesFile.eof()) {
    getline( statesFile, board );
    if (board != "")
      stateList.push_back(board);
  }
  statesFile.close();
  return stateList.size();
}


int greedy(const string& board, char player, map<string, double>& stateValue) {
  string nextBoard         = board;
  double maxValue          = -1000.0;
  int    bestMove          =    -1;
  int    MOVES[BOARD_SIZE] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15};
  //random_shuffle( MOVES, MOVES + 9 );
  for (int i = 0; i < BOARD_SIZE; ++i) {
    int move = MOVES[i];
    if (isLegal(move, nextBoard)) {
      nextBoard[move] = player;
      const double v = stateValue[TRANSFORM[nextBoard]];
      if (v > maxValue) {
        maxValue = v;
        bestMove = move;
      }
      nextBoard[move] = EMPTY;
    }
  }
  return bestMove;
}



int e_greedy(const string& board, char player, map<string, double>& stateValue, double epsilon )
{
  if ((double(rand()) / RAND_MAX) < epsilon) {
    int move = rand() % BOARD_SIZE;
    while (!isLegal( move, board))
      move = rand() % BOARD_SIZE;
    return move;
  }
  return greedy(board, player, stateValue);
}



int computerMoveOptimal(const string& board, char computer, map<string, eState>& stateInfo) {
  double S                 = (computer == X ? 1.0 : -1.0);
  string nextBoard         = board;
  double maxValue          = -1000.0;
  int    bestMove          =    -1;
  int    MOVES[BOARD_SIZE] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15};
  random_shuffle( MOVES, MOVES + BOARD_SIZE );
  for (int i = 0; i < BOARD_SIZE; ++i) {
    int move = MOVES[i];
    if (isLegal(move, nextBoard)) {
      nextBoard[move] = computer;
      const double v = S * stateInfo[nextBoard].value;
      if (v > maxValue) {
        maxValue = v;
        bestMove = move;
      }
      nextBoard[move] = EMPTY;
    }
  }
  return bestMove;
}



void initValueFunction(const vector<string> stateList, map<string, double>& stateValue, double min, double max) {
  for (vector<string>::const_iterator ci = stateList.begin(); ci < stateList.end(); ++ci) {
    const string& board = *ci;
    if (winner(board) != NO_ONE)
      stateValue[board] = 0.0;
    else
      stateValue[board] = double(rand()) / RAND_MAX * (max - min) + min;
  }
}



double reward(const string& board, char player) {
  if (winner(board) == player)
    return 1.0;
  else if (winner(board) == opponent(player))
    return -1.0;
  return 0.0;
}


void td_update(map<string, double>& stateValue, const string& oldState, const string& newState, double reward, double alpha, double gamma) {
  const double V_t   = stateValue[oldState];
  const double V_t1  = stateValue[newState];
  const double delta = reward + gamma * V_t1 - V_t;
  stateValue[oldState] = V_t + alpha * delta;
}



void experiment(const vector<string> stateList, map<string, eState>& stateInfo, int l, int t, int z, double alpha, double eps, double gamma) {
  cerr << "\n* Experiment Parameters: L=" << l << "  ;  T=" << t << "  ;  Z="  << z << "\n";
  cerr << "* Learning Parameters  : alpha=" << alpha << "  ;  epsilon=" << eps << "  ;  gamma="  << gamma << "\n";

  int lost = 0;
  int tie  = 0;
  int won  = 0;

  map<string, double> stateValue;
  map<string, double> e_trace;

  for (int exp = 1; exp <= z; ++exp) { // aeussere Experiment-Schleife
    initValueFunction( stateList, stateValue, 0.0, 0.0 );

    char agent = X;     // Lernagent spielt mit X
    char opp   = opponent(agent);
    char turn  = X;   // X faengt an
    int move;
    string board(BOARD_SIZE, EMPTY); // leeres Feld erzeugen

    for (int learn = 1; learn <= l; ++learn) { // Lernschleife
      e_trace.clear();
      turn  = X;
      board = string(BOARD_SIZE, EMPTY);
      while (winner(board) == NO_ONE) { // ab hier wird gespielt
        const string oldBoard = board;
        if (turn == opp) {
          move        = computerMoveOptimal(board, opp, stateInfo);
          board[move] = opp;
          const double r = reward(board, agent);
          td_update(stateValue, TRANSFORM[oldBoard], TRANSFORM[board], r, alpha, gamma);
        } else {
          move        = e_greedy(board, agent, stateValue, eps);
          board[move] = agent;
          const double r = reward(board, agent);
          td_update(stateValue, TRANSFORM[oldBoard], TRANSFORM[board], r, alpha, gamma);
        }
        turn = opponent(turn);
      }
    }

    for (int test = 1; test <= t; ++test) {  // ab hier wird getestet
      turn  = X;
      board = string(BOARD_SIZE, EMPTY);
      while (winner(board) == NO_ONE) {
        if (turn == opp) {
          move        = computerMoveOptimal(board, opp, stateInfo);
          board[move] = opp;
        } else {
          move        = greedy(board, agent, stateValue);
          board[move] = agent;
        }
        turn = opponent(turn);
      }
      if (winner(board) == agent)
	won++;
      if (winner(board) == opp)
	lost++;
      if (winner(board) == TIE) 
	tie++;
    }
  }
  cerr << "\n>>> Test Stat : LOST = " << lost << "  |  TIE = " << tie << "  |  WON = "  << won << "\n";
}



string altState(const string& board) {
  int aState = 0; // das wird unser neuer Zustand
  const int WINNING_ROWS[10][4] = {{0, 1, 2,  3}, {4, 5,  6,  7}, {8, 9, 10, 11}, {12, 13, 14, 15}, {0, 4, 8, 12},
				   {1, 5, 9, 13}, {2, 6, 10, 14}, {3, 7, 11, 15}, { 0,  5, 10, 15}, {3, 6, 9, 12}};
  for (int i = 0; i < 10; i++) {
    int k = 0;
    int l = 0;
    for (int j = 0; j < 4; j++)
      if (board[WINNING_ROWS[i][j]] == X)
        ++k;
      else if (board[WINNING_ROWS[i][j]] == O)
        ++l;
    if (k == 1)
      ++aState;
    else if (k == 2)
      aState += 11;
    else if (k == 3)
      aState += 121;
    else if (k == 4)
      return "X";
    else if (l == 1)
      aState += 1331;
    else if (l == 2)
      aState += 14641;
    else if (l == 3)
      aState += 161051;
    else if (l == 4)
      return "O";
  }
  // der Kompatibilitt mit vorhandenen Funktionen halber konvertieren wir den neuen Zustandsreprsentanten wieder in einen string
  ostringstream s;
  s << aState;
  return s.str();
}
